"""Module for map-matching source linestrings back to the topological graph."""

import math
from itertools import count
from heapq import heappush, heappop
import networkx as nx
from rtree import index
from shapely.geometry import Point
import numpy as np
import logging

logger = logging.getLogger('network-topology')
logger.setLevel(logging.INFO)


def getMatchedRoutes(lineStrings, DG, increment=100., scope=30.):
    """List the segments/edges that best match each linestring."""
    idx = _buildIndex(DG)
    paths = {}
    for i, ls in lineStrings.items():
        logger.info('Matching shape {}'.format(i))
        adjacency_list, s, t = _findCandidatePoints(DG, idx, ls,
                                                    increment=increment,
                                                    scope=scope)
        path = _viterbi_search(adjacency_list, s, t, DG, increment)
        paths[i] = path
    return paths


def _buildIndex(DG):
    shapeIdx = index.Index()
    for u, v, i, data in DG.edges(data=True, keys=True):
        shapeIdx.insert(hash((u, v, i)), data['geom'].bounds, obj=(u, v, i))
    return shapeIdx


def _getDirection(line, distance):
    if distance == line.length:
        n1 = 1. - 1e-9
        n2 = 1.
    else:
        n1 = distance / line.length
        n2 = n1 + 1e-9
    p1 = line.interpolate(n1, normalized=True)
    p2 = line.interpolate(n2, normalized=True)
    return (p2.x - p1.x, p2.y - p1.y)


def _findCandidatePoints(DG, shapeIdx, lineString, increment=100., scope=30.):
    adjacency_list = {}
    totalLength = lineString.length
    s = Candidate(DG, start=True)
    t = Candidate(DG, end=True)
    lastRound = [s]
    adjacency_list[t] = []

    simpleLine = lineString.simplify(scope/2.)
    cornerPts = [Point(c) for c in simpleLine.coords[1:-1]]

    points = [(Point(lineString.coords[0]), _getDirection(lineString, 0))]
    searchRange = [increment, min(increment*2.0, totalLength - increment)]
    failureCount = 0
    while searchRange[1] > searchRange[0]:
        d = np.random.uniform(*searchRange)
        p = lineString.interpolate(d)
        collision = False
        for corner in cornerPts:
            if abs(p.x - corner.x) < scope and abs(p.y - corner.y) < scope:
                if p.distance(corner) < scope:
                    failureCount += 1
                    collision = True
                    break
        if failureCount >= 20:
            searchRange = [searchRange[1],
                           min(searchRange[1]+scope, totalLength-increment)]
            failureCount = 0
        if not collision:
            points.append((p, _getDirection(lineString, d)))
            searchRange = [d + increment,
                           min(d + 2. * increment, totalLength-increment)]
    points.append((Point(lineString.coords[-1]),
                   _getDirection(lineString, totalLength)))

    firstLast = points[::len(points)-1]
    for p, pdir in points:
        nextRound = []
        for item in shapeIdx.intersection((p.x-scope, p.y-scope,
                                           p.x+scope, p.y+scope),
                                          objects=True):
            u, v, i = item.object
            if DG[u][v][i]['terminal'] and p not in firstLast:
                continue
            candidate = Candidate(DG, segment=item.object, measurement=p)
            if candidate.distance < scope:
                sdir = _getDirection(DG[u][v][i]['geom'], candidate.offset)
                if np.dot(pdir, sdir) >= 0:
                    nextRound.append(candidate)
        if nextRound:
            for prev in lastRound:
                adjacency_list[prev] = list(nextRound)
            lastRound = nextRound
            #pprint(nextRound)
    for prev in lastRound:
        adjacency_list[prev] = [t]
    return adjacency_list, s, t


def _viterbi_search(adjacency_list, s, t, DG, increment):
    # With credit to the "Map Matching in a Programmer's Perspective" guide
    # in Valhalla by Mapzen:
    # https://github.com/valhalla/meili/blob/master/docs/meili/algorithms.md

    # Initialize joint probability for each node
    if 'path_cache' not in DG.graph:
        DG.graph['path_cache'] = SequenceCache()
    cost = {}
    for u in adjacency_list:
        cost[u] = 1e100
    predecessor = {}
    queue = PriorityQueue()

    cost[s] = s.emission_cost()
    queue.add_or_update(s, cost[s])
    predecessor[s] = None
    while not queue.empty():
        # Extract node u
        u = queue.pop()
        if u == t:
            break
        for v in adjacency_list[u]:
            # Relaxation
            new_cost = (cost[u] +
                        u.transition_cost(v, DG.graph['path_cache']) +
                        v.emission_cost())
            if cost[v] > new_cost:
                cost[v] = new_cost
                predecessor[v] = u
            queue.add_or_update(v, cost[v])
    return _construct_path(predecessor, s, t, DG, increment)


def _construct_path(predecessor, s, t, DG, increment):
    cache = DG.graph['path_cache']
    cur = predecessor[t]
    sequence = [cur.segment]
    prev = predecessor[cur]
    while prev is not s:
        if not(prev.segment == cur.segment and prev.offset <= cur.offset):
            if (prev.segment[1], cur.segment[0]) in cache:
                cVal = cache[prev.segment[1], cur.segment[0]][1]
                sequence = cVal + sequence
        if (len(sequence) == 0 or sequence[0] != prev.segment):
            sequence.insert(0, prev.segment)
        cur = prev
        prev = predecessor[cur]
    sequence = [(u, v, i) for u, v, i in sequence if not DG[u][v][i]['terminal']]

    duplicates = []
    for s1, s2 in zip(sequence[:-1], sequence[1:]):
        if s1[0] == s2[1] and s1[2] == s2[2]:
            print s1, s2, DG[s1[0]][s1[1]][s1[2]]['geom'].length
            if DG[s1[0]][s1[1]][s1[2]]['geom'].length < increment*2.:
                duplicates.append(s1)
                duplicates.append(s2)

    return [p for p in sequence if p not in duplicates]


def _distanceBetweenNodes(G, a, b):
    x1 = G.node[a]['x']
    y1 = G.node[a]['y']
    x2 = G.node[b]['x']
    y2 = G.node[b]['y']
    return math.sqrt((x1-x2)**2. + (y1-y2)**2.)


def _astar_path(G, source, target, weight='length'):
    """Return a list of edges in a shortest path between source and target
    using the A* ("A-star") algorithm.
    There may be more than one shortest path.  This returns only one.
    This code is adapted for multigraph from the networkx library
    Parameters
    ----------
    G : NetworkX graph
    source : node
       Starting node for path
    target : node
       Ending node for path
    heuristic : function
       A function to evaluate the estimate of the distance
       from the a node to the target.  The function takes
       two nodes arguments and must return a number.
    weight: string, optional (default='weight')
       Edge data key corresponding to the edge weight.
    Raises
    ------
    NetworkXNoPath
        If no path exists between source and target.
    Examples
    --------
    >>> G = nx.path_graph(5)
    >>> print(nx.astar_path(G, 0, 4))
    [0, 1, 2, 3, 4]
    >>> G = nx.grid_graph(dim=[3, 3])  # nodes are two-tuples (x,y)
    >>> nx.set_edge_attributes(G, {e: e[1][0]*2 for e in G.edges()}, 'cost')
    >>> def dist(a, b):
    ...    (x1, y1) = a
    ...    (x2, y2) = b
    ...    return ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5
    >>> print(nx.astar_path(G, (0, 0), (2, 2), heuristic=dist, weight='cost'))
    [(0, 0), (0, 1), (0, 2), (1, 2), (2, 2)]
    See Also
    --------
    shortest_path, dijkstra_path
    """
    if source not in G or target not in G:
        msg = 'Either source {} or target {} is not in G'
        raise nx.NodeNotFound(msg.format(source, target))

    # if heuristic is None:
    #     # The default heuristic is h=0 - same as Dijkstra's algorithm
    #     def heuristic(u, v):
    #         return 0

    push = heappush
    pop = heappop

    # The queue stores priority, node, cost to reach, and parent.
    # Uses Python heapq to keep in priority order.
    # Add a counter to the queue to prevent the underlying heap from
    # attempting to compare the nodes themselves. The hash breaks ties in the
    # priority and is guarenteed unique for all nodes in the graph.
    c = count()
    queue = [(0, next(c), source, 0, None, None)]

    # Maps enqueued nodes to distance of discovered paths and the
    # computed heuristics to target. We avoid computing the heuristics
    # more than once and inserting the node into the queue too many times.
    enqueued = {}
    # Maps explored nodes to parent closest to the source.
    explored = {}

    while queue:
        # Pop the smallest item from queue.
        _, __, curnode, dist, parent, via = pop(queue)

        if curnode == target:
            path = []
            node = curnode
            edge = via
            while parent is not None:
                path.append((parent, node, edge))
                node = parent
                parent, edge = explored[node]
            path.reverse()
            return path

        if curnode in explored:
            continue

        explored[curnode] = parent, via

        for neighbor, es in G[curnode].items():
            for edge, w in es.items():
                if neighbor in explored:
                    continue
                ncost = dist + w.get(weight, 1)
                if neighbor in enqueued:
                    qcost, h = enqueued[neighbor]
                    # if qcost < ncost, a longer path to neighbor remains
                    # enqueued. Removing it would need to filter the whole
                    # queue, it's better just to leave it there and ignore
                    # it when we visit the node a second time.
                    if qcost <= ncost:
                        continue
                else:
                    h = _distanceBetweenNodes(G, neighbor, target)
                enqueued[neighbor] = ncost, h
                push(queue, (ncost + h, next(c), neighbor, ncost, curnode, edge))

    raise nx.NetworkXNoPath("Node %s not reachable from %s" % (source, target))

SIGMA_Z = 4.07
BETA = 3
INVERSE_2_SIGMA_SQUARED = 1. / (SIGMA_Z ** 2. * 2.)
INVERSE_BETA = 1. / BETA

class Candidate:
    """Class for candidate segment."""

    def __init__(self, DG, segment=None, measurement=None,
                 start=False, end=False):
        """Initialize."""
        self.start = start
        self.end = end
        self.measurement = measurement
        self.segment = segment
        self.DG = DG
        if segment:
            geom = DG.get_edge_data(*segment)['geom']
            self.offset = geom.project(measurement)
            self.distance = geom.distance(measurement)
            self.remaining = geom.length - self.offset
            self.hash = hash((self.segment, self.offset))
        else:
            self.distance = 0
            self.offset = 0
            self.remaining = 0
            if self.start:
                self.hash = hash('start')
            else:
                self.hash = hash('end')


    def __hash__(self):
        """Hash for use as dictionary keys."""
        return self.hash

    def __str__(self):
        """String representation as segment id and offset."""
        return '{}>{}'.format(self.segment, self.offset)

    def __repr__(self):
        """Just use string representation."""
        return self.__str__()

    def emission_cost(self):
        """Gaussian distribution."""
        #c = 1 / (SIGMA_Z * math.sqrt(2 * math.pi))
        #prob = c * math.exp(-(self.distance**2))
        #return -1 * math.log10(max(prob, sys.float_info.min))
        return self.distance ** 2. * INVERSE_2_SIGMA_SQUARED

    def transition_cost(self, nextCandidate, cache):
        """Empirical distribution."""
        if self.measurement is None:
            return nextCandidate.offset
        elif nextCandidate.measurement is None:
            return self.remaining
        #c = 1 / BETA
        route_distance, crossed = self._route_distance_to(nextCandidate, cache)
        delta = abs(route_distance -
                    self.measurement.distance(nextCandidate.measurement))
        # Apply a small bias against changing edges, to avoid swithcing between
        # opposide edges unnecessarily.
        #delta *= 1.000000000001**crossed
        delta += .001 * crossed

        #prob = c * math.exp(-delta*c)
        #return -1 * math.log10(max(prob, sys.float_info.min))
        return delta * INVERSE_BETA

    def _route_distance_to(self, nextCandidate, cache):
        if nextCandidate.segment == self.segment and nextCandidate.offset >= self.offset:
            return nextCandidate.offset - self.offset, 0
        if self.segment is None or nextCandidate.segment is None:
            return 0, 0
        terminalDistance = self.remaining + nextCandidate.offset
        if self.segment[1] == nextCandidate.segment[0]:
            return terminalDistance, 1
        sumDistance = 0
        index = (self.segment[1], nextCandidate.segment[0])
        if index in cache:
            sumDistance, edgeSequence = cache[index]
        else:
            edgeSequence = []
            distances = []
            try:
                sequence = _astar_path(self.DG, self.segment[1],
                                            nextCandidate.segment[0],
                                            weight='length')
                #for u, v in zip(sequence[:-1], sequence[1:]):
                #    i = min(self.DG[u][v],
                #            key=lambda x: self.DG[u][v][x]['length'])
                for u, v, i in sequence:
                    edgeSequence.append((u, v, i))
                    distances.append(self.DG[u][v][i]['length'])
            except nx.NetworkXException:
                distances = [float('inf')]
            #cache[index] = sumDistance, edgeSequence
            cache.add(distances, edgeSequence)
        return sumDistance + terminalDistance, len(edgeSequence) + 1


class PriorityQueue:
    """Priority queue for use in Dijkstra search."""

    REMOVED = '<removed-item>'  # placeholder for a removed item

    def __init__(self):
        """Initialize priority queue."""
        # self.pq = Fibonacci_heap()
        # self.entry_finder = {}
        self.pq = []                       # list of entries arranged in a heap
        self.entry_finder = {}             # mapping of items to entries
        self.counter = count()   # unique sequence count

    def add_or_update(self, item, priority=0):
        """Add a new item or update the priority of an existing item."""
        # if item in self.entry_finder:
        #     self.pq.decrease_key(self.entry_finder[item], priority)
        # else:
        #     entry = self.pq.enqueue(item, priority)
        #     self.entry_finder[item] = entry
        if item in self.entry_finder:
            self.remove(item)
        count = next(self.counter)
        entry = [priority, count, item]
        self.entry_finder[item] = entry
        heappush(self.pq, entry)

    def remove(self, item):
        """Mark an existing item as REMOVED.  Raise KeyError if not found."""
        entry = self.entry_finder.pop(item)
        entry[-1] = self.REMOVED

    def pop(self):
        """Remove and return the lowest priority task. Raise error if empty."""
        while self.pq:
            priority, count, item = heappop(self.pq)
            if item is not self.REMOVED:
                del self.entry_finder[item]
                return item
        raise KeyError('pop from an empty priority queue')
        # item = self.pq.dequeue_min().get_value()
        # return item

    def empty(self):
        """Return true if the priority queue is empty."""
        return len(self.pq) == 0


class SequenceCache(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __contains__(self, key):
        s, t = key
        return dict.__contains__(self, (s, t)) or dict.__contains__(self, (t, s))

    def __getitem__(self, key):
        s, t = key
        if dict.__contains__(self, (s, t)):
            return dict.__getitem__(self, key)
        elif dict.__contains__(self, (t, s)):
            length, sequence = dict.__getitem__(self, (t, s))
            sequence = [(v, u, i) for u, v, i in reversed(sequence)]
            return length, sequence
        else:
            raise KeyError(key)

    def add(self, lengths, sequence):
        s, t = sequence[0][0], sequence[-1][1]
        dict.__setitem__(self, (s, t), (sum(lengths), sequence))
        if len(sequence) > 1:
            s, t = sequence[0][0], sequence[-2][1]
            if (s, t) not in self:
                self.add(lengths[:-1], sequence[:-1])
            s, t = sequence[1][0], sequence[-1][1]
            if (s, t) not in self:
                self.add(lengths[1:], sequence[1:])

    # def __setitem__(self, key, sequence):
    #     dict.__setitem__(self, key, sequence)
    #     if len(sequence) > 1:
    #         if

    def __repr__(self):
        dictrepr = dict.__repr__(self)
        return '{}({})'.format(type(self).__name__, dictrepr)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).iteritems():
            self[k] = v



"""Module for creating a simplified graph from a set of linestrings."""

import networkx as nx
from triangle import triangulate
from shapely.geometry import LineString, Point, Polygon, MultiPolygon, mapping
from shapely.ops import cascaded_union
from itertools import combinations
import numpy as np
import random
from rtree import index
import logging
import json
import pyproj
from shapely.ops import transform
from functools import partial
import os

logger = logging.getLogger('network-topology')
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
logger.addHandler(ch)


def getNetworkTopology(lineStrings, thickness=14.0, splitAtTeriminals=None,
                       turnThreshold=20.0, minInnerPerimeter=200,
                       debugFolder=None):
    """Generate a bidirectional graph of the topology created by the lines."""
    # Make buffered shape
    big_shape = _makeBufferedShape(lineStrings, thickness, minInnerPerimeter)
    if debugFolder:
        _dumpBigShape(big_shape, debugFolder)

    triangles = _triangulate(big_shape, lineStrings, thickness,
                             minInnerPerimeter)
    if debugFolder:
        _dumpTriangles(triangles, debugFolder)

    polygons = _polygonize(triangles, big_shape)
    if debugFolder:
        _dumpPolygons(polygons, triangles, debugFolder)

    poly_graph = _buildInitialGraph(polygons)
    _tidyIntersections(poly_graph, triangles, polygons)
    skel_graph = _buildSkeletonGraph(poly_graph, triangles, polygons)
    if splitAtTeriminals:
        _insertAtTerminals(skel_graph, splitAtTeriminals)
    graph = _simplifyGraph(skel_graph, turnThreshold, thickness/100.0)
    # _findSegmentsAndIntersections(skel_graph, turnThreshold)
    return graph


def _makeBufferedShape(lineStrings, thickness=14.0, minInnerPerimeter=200):
    logger.info('Creating buffered shape')
    bufferedShapes = []
    buf = thickness / 2.0
    for ls in lineStrings:
        bufferedShapes.append(ls.buffer(buf, resolution=2, join_style=3))
    big_shape = cascaded_union(bufferedShapes)
    if big_shape.geom_type == 'MultiPolygon':
        filled_shape = MultiPolygon([[g.exterior.coords,
                                     [ring.coords for ring in g.interiors
                                      if ring.length > minInnerPerimeter]]
                                     for g in big_shape.geoms])
    else:
        filled_shape = Polygon(big_shape.exterior.coords,
                               [ring.coords for ring in big_shape.interiors
                                if ring.length > minInnerPerimeter])
    bs = filled_shape.simplify(thickness/10.0)
    logger.info('Completed buffered shape')
    return bs


def _triangulate(big_shape, lineStrings, thickness=14.0, minInnerPerimeter=200):
    logger.info('Reticulating shape')
    tri = {
        'vertices': [],
        'segments': []
        }
    if big_shape.geom_type == 'MultiPolygon':
        geoms = big_shape.geoms
    else:
        geoms = [big_shape]
    for shape in geoms:
        lastIndex = len(tri['vertices'])
        tri['vertices'].append(list(shape.exterior.coords[0]))
        for x, y in shape.exterior.coords[1:]:
            if [x, y] in tri['vertices']:
                currentIndex = tri['vertices'].index([x, y])
            else:
                currentIndex = len(tri['vertices'])
                tri['vertices'].append([x, y])
            tri['segments'].append([lastIndex, currentIndex])
            lastIndex = currentIndex

        if shape.interiors and 'holes' not in tri:
            tri['holes'] = []
        for ring in shape.interiors:
            if ring.length > minInnerPerimeter:
                x, y = 0, 0
                minx, miny, maxx, maxy = ring.bounds
                while not Polygon(ring).contains(Point(x, y)):
                    x = random.uniform(minx, maxx)
                    y = random.uniform(miny, maxy)
                tri['holes'].append([x, y])
                lastIndex = len(tri['vertices'])
                tri['vertices'].append(list(ring.coords[0]))
                for x, y in ring.coords[1:]:
                    if [x, y] in tri['vertices']:
                        currentIndex = tri['vertices'].index([x, y])
                    else:
                        currentIndex = len(tri['vertices'])
                        tri['vertices'].append([x, y])
                    tri['segments'].append([lastIndex, currentIndex])
                    lastIndex = currentIndex

    for ls in lineStrings:
        if ls.geom_type == 'LineString':
            geoms = [ls]
        else:
            geoms = ls.geoms
        for geom in geoms:
            for i in [0, -1]:
                coords = list(geom.coords[i])
                if coords not in tri['vertices']:
                    tri['vertices'].append(coords)
    tri = triangulate(tri, 'pq0Da{}i'.format(thickness**2.0))
    logger.info('Reticulating complete')
    return tri


def _polygonize(res, big_shape):
    logger.info('Removing internal nodes')
    interior_pts = {}
    for i, v in enumerate(res['vertex_markers']):
        if v == 0:
            interior_pts[i] = []
    for tri, pts in enumerate(res['triangles']):
        for p in pts:
            if p in interior_pts:
                interior_pts[p].append(tri)
    connectivity_graph = nx.Graph()
    for tris in interior_pts.values():
        if len(tris) > 0:
            for tri in tris[1:]:
                connectivity_graph.add_edge(tris[0], tri)
    polys = []
    for cluster in list(nx.connected_components(connectivity_graph)):
        shape_graph = nx.Graph()
        for tri in cluster:
            for v1, v2 in zip(res['triangles'][tri],
                              list(res['triangles'][tri][1:]) +
                              [res['triangles'][tri][0]]):
                if v1 not in interior_pts and v2 not in interior_pts:
                    shape_graph.add_edge(v1, v2)

        poly = []
        pt = shape_graph.nodes().keys()[0]
        while pt not in poly:
            poly.append(pt)
            for neighbour in shape_graph.neighbors(pt):
                if neighbour not in poly and len(shape_graph[neighbour]) != 1:
                    pt = neighbour
                    break
        polys.append(poly)
    remaining_triangles = [list(vs) for i, vs in enumerate(res['triangles'])
                           if i not in connectivity_graph]
    return remaining_triangles + polys


def _mergeShapes(cluster, TG, polys):
    combined_graph = nx.Graph()
    for shape in cluster:
        if len(polys[shape]) == 0:
            continue
        for p1, p2 in zip(polys[shape], polys[shape][1:] + [polys[shape][0]]):
            if combined_graph.has_edge(p1, p2):
                combined_graph.remove_edge(p1, p2)
                if not combined_graph[p1]:
                    combined_graph.remove_node(p1)
                if not combined_graph[p2]:
                    combined_graph.remove_node(p2)
            else:
                combined_graph.add_edge(p1, p2)
    poly = []
    pt = combined_graph.nodes().keys()[0]
    while pt not in poly:
        poly.append(pt)
        for neighbour in combined_graph.neighbors(pt):
            if neighbour not in poly:
                pt = neighbour
                break
    cluster = list(cluster)
    base = cluster[0]
    polys[base] = poly
    for shape in cluster[1:]:
        polys[shape] = []
        if shape in TG:
            for n, d in TG[shape].iteritems():
                if n != base:
                    TG.add_edge(base, n, **d)
            TG.remove_node(shape)


def _buildInitialGraph(polys):
    logger.info('Build polygon graph')
    edge_tri = {}
    TG = nx.Graph()

    def add_edge(n1, n2, tri, edges, TG):
        n1, n2 = sorted([n1, n2])
        if (n1, n2) not in edges:
            edges[n1, n2] = []
        for other in edges[n1, n2]:
            TG.add_edge(tri, other, v1=n1, v2=n2)
        edges[n1, n2].append(tri)

    for i, t in enumerate(polys):
        for t1, t2 in zip(t[:-1], t[1:]):
            add_edge(t1, t2, i, edge_tri, TG)
        add_edge(t[0], t[-1], i, edge_tri, TG)

    # Collapse internal edges
    central_shapes = [t for t in TG.nodes() if len(TG[t]) > 2]
    SG = TG.subgraph(central_shapes)
    clusters = list(nx.connected_components(SG))

    for cluster in clusters:
        _mergeShapes(cluster, TG, polys)

    logger.info('Polygon graph complete')
    return TG


def _tidyIntersections(TG, res, polys):
    logger.info('Tyding intersections')
    # 'Tyding T-intersections...'
    toMerge = []
    for node in TG.nodes():
        if TG.degree(node) == 3 and len(polys[node]) == 3:
            points = res['vertices'][polys[node]]
            A = np.array([[p[0]*p[0] + p[1]*p[1], p[0], p[1], 1]
                          for p in points])
            M11 = np.linalg.det(A[:, [1, 2, 3]])
            M12 = np.linalg.det(A[:, [0, 2, 3]])
            M13 = np.linalg.det(A[:, [0, 1, 3]])
            M14 = np.linalg.det(A[:, [0, 1, 2]])
            x = 0.5 * M12/M11
            y = -0.5 * M13/M11
            r2 = M14/M11 + x*x + y*y
            TG.node[node]['centroid'] = [x, y]

            for neighbour, data in TG[node].iteritems():
                points = list(polys[neighbour])
                points.remove(data['v1'])
                points.remove(data['v2'])
                point = res['vertices'][points[0]]
                dx = x - point[0]
                dy = y - point[1]
                if dx*dx + dy*dy < 4*r2:
                    toMerge.append([node, neighbour])

    for cluster in toMerge:
        _mergeShapes(cluster, TG, polys)

    logger.info('Tyding complete')
    # 'Cleaning up angled crossings...'
    # toMerge = []
    # for node in TG.nodes_iter():
    #     if TG.degree(node) == 2:
    #         sandwiched = True
    #         for neighbour in TG[node]:
    #             if TG.degree(neighbour) == 2:
    #                 sandwiched = False
    #         if sandwiched:
    #             toMerge.append([node] + TG[node].keys())
    #
    # for cluster in toMerge:
    #     _mergeShapes(cluster, TG, polys)


def _makeKey(t1, t2):
    t1, t2 = sorted([t1, t2])
    return int(t1 * 1e9 + t2)


def _normalVector(G, n1, n2):
    ux = G.node[n1]['x'] - G.node[n2]['x']
    uy = G.node[n1]['y'] - G.node[n2]['y']
    dist = max(np.sqrt(ux*ux + uy*uy), 1e-18)
    ux /= dist
    uy /= dist
    return ux, uy


def _insertEdge(G, n1, n2, terminal=False):
    if n1 == n2:
        return
    ux, uy = _normalVector(G, n1, n2)
    G.add_edge(n1, n2, ux=ux, uy=uy, terminal=terminal)


def _buildSkeletonGraph(TG, res, polys):
    logger.info('Building skeleton graph')
    node_idx = index.Index()
    G = nx.Graph(index=node_idx)
    for t1, t2, d in TG.edges(data=True):
        x = (res['vertices'][d['v1']][0] + res['vertices'][d['v2']][0])/2.0
        y = (res['vertices'][d['v1']][1] + res['vertices'][d['v2']][1])/2.0
        key = _makeKey(t1, t2)
        node_idx.insert(key, (x, y, x, y))
        G.add_node(key, x=x, y=y)

    for t, d in TG.nodes(data=True):
        if len(TG[t]) == 2:
            pairs = []
            for t2 in TG[t]:
                key = _makeKey(t, t2)
                if key in G:
                    pairs.append(key)
            for e1, e2 in combinations(pairs, 2):
                _insertEdge(G, e1, e2)
        else:  # if len(TG[t]) != 2:
            if 'centroid' in TG.node[t]:
                x, y = TG.node[t]['centroid']
            elif len(polys[t]) > 2:
                # calc centroid & add to graph
                poly = Polygon([list(res['vertices'][v])
                                for v in polys[t]])
                x, y = poly.centroid.coords[0]
            else:
                x, y = zip(*[list(res['vertices'][v])
                             for v in polys[t]])
                x = sum(x)/len(x)
                y = sum(y)/len(y)

            G.add_node(t, x=x, y=y)
            node_idx.insert(t, [x, y, x, y])
            # join all edges to centroid
            for t2 in TG[t]:
                key = _makeKey(t, t2)
                _insertEdge(G, t, key)
    logger.info('Skeleton graph complete')
    return G


def _insertAtTerminals(G, points):
    logger.info('Inserting terminal nodes')
    node_idx = G.graph['index']
    terminals = []
    #for lid, ls in enumerate(lineStrings):
    #    for i in [0, -1]:
    #        point = ls.coords[i]
    for i, point in enumerate(points):
            nearest = list(node_idx.nearest(point.bounds, num_results=1))[0]
            pid = -float(i) #-float(lid) + 0.1 * i
            terminals.append((pid, point, nearest))
    for pid, point, nearest in terminals:
        G.add_node(pid, x=point.x, y=point.y, terminal=True)
        _insertEdge(G, pid, nearest, terminal=True)


def _simplifyGraph(G, turnThreshold=20.0, simplifyTolerance=0):
    logger.info('Simplifying graph')
    DG = nx.MultiDiGraph()
    threshold = np.cos(turnThreshold/180.0*np.pi)
    for node in G.nodes():
        if node not in DG:
            DG.add_node(node, x=G.node[node]['x'], y=G.node[node]['y'])
        else:
            DG.node[node]['x'] = G.node[node]['x']
            DG.node[node]['y'] = G.node[node]['y']
        if len(G[node]) == 2:
            normals = []
            for neighbour in G[node]:
                n = _normalVector(G, node, neighbour)
                normals.append(n)
            dp = -(normals[0][0] * normals[1][0] +
                   normals[0][1] * normals[1][1])
            DG.node[node]['turn'] = dp < threshold
        if 'terminal' in G.node[node]:
            DG.node[node]['terminal'] = True
        for neighbour in G[node]:
            DG.add_edge(node, neighbour, nnodes=[node, neighbour],
                        terminal=G[node][neighbour]['terminal'])
    to_remove = []
    for node in DG.nodes():
        if len(DG.edges(node)) == 2 and not DG.node[node]['turn'] and node not in DG.neighbors(node):
            (x, n1), (x, n2) = DG.edges(node)
            if DG[node][n1][0]['terminal'] or DG[node][n2][0]['terminal']:
                continue
            seq = DG[n1][node][0]['nnodes'] + DG[node][n2][0]['nnodes'][1:]
            DG.add_edge(n1, n2, nnodes=seq, terminal=False)
            DG.add_edge(n2, n1, nnodes=list(reversed(seq)), terminal=False)
            to_remove.append(node)
            DG.remove_edge(n1, node)
            DG.remove_edge(node, n1)
            DG.remove_edge(n2, node)
            DG.remove_edge(node, n2)
        #if 'terminal' in DG.node[node]:
        #    to_remove.append(node)
    for node in to_remove:
        DG.remove_node(node)
    for s, t, data in DG.edges(data=True):
        coords = []
        for node in data['nnodes']:
            coords.append((G.node[node]['x'], G.node[node]['y']))
        ls = LineString(coords).simplify(simplifyTolerance)
        data['geom'] = ls
        data['length'] = ls.length
    for node, data in DG.nodes(data=True):
        data['geom'] = Point(G.node[node]['x'], G.node[node]['y'])
    logger.info('Simplificaiton complete')
    return DG


def _dumpBigShape(big_shape, debugFolder):
    zone = (int((big_shape.exterior.coords[0][0] + 180)/6) % 60) + 1
    utm17 = pyproj.Proj(proj='utm', zone=zone, ellps='WGS84')
    google = pyproj.Proj(init='epsg:4326')
    project = partial(pyproj.transform, utm17, google)
    with open(os.path.join(debugFolder, 'bigshape.geojson'), 'w') as fp:
        json.dump(mapping(transform(project, big_shape)), fp)


def _dumpTriangles(triangles, debugFolder):
    zone = (int((triangles['vertices'][0][0] + 180)/6) % 60) + 1
    utm17 = pyproj.Proj(proj='utm', zone=zone, ellps='WGS84')
    features = []
    for vs in triangles['triangles']:
        coord_list = [list(utm17(inverse=True,
                                 *triangles['vertices'][v])) for v in vs]
        features.append({"type": "Feature",
                         "properties": {'stroke-opacity': 0.25},
                         "geometry": {
                            "type": "Polygon",
                            "coordinates": [coord_list]
                         }})
    data = {"type": "FeatureCollection",
            "features": features}
    with open(os.path.join(debugFolder, 'triangles.geojson'), 'w') as fp:
        json.dump(data, fp)


def _dumpPolygons(polygons, triangles, debugFolder):
    zone = (int((triangles['vertices'][0][0] + 180)/6) % 60) + 1
    utm17 = pyproj.Proj(proj='utm', zone=zone, ellps='WGS84')
    features = []
    for vs in polygons:
        coord_list = [list(utm17(inverse=True,
                      *triangles['vertices'][v])) for v in vs]
        if not coord_list:
            continue
        features.append({"type": "Feature",
                         "properties": {'stroke-opacity': 0.25},
                         "geometry": {
                            "type": "Polygon",
                            "coordinates": [coord_list]
                         }})
    data = {"type": "FeatureCollection",
            "features": features}
    with open(os.path.join(debugFolder, 'polygons.geojson'), 'w') as fp:
        json.dump(data, fp)
